import datetime, os, re
import boto3
from botocore.exceptions import ClientError
from pathlib import Path
from pyawscp.Config import Config
from pyawscp.PrettyTable import PrettyTable
from pyawscp.Emoticons import Emoticons
from pyawscp.TableArgs import TableArgs
from pyawscp.Utils import Utils, Style
from pyawscp.messages import Messages
from pyawscp.Functions import Functions

class AwsSession:

    def __init__(self, config):
        self.credentials = {}
        self.config      = config
        self.parse_credentials()
    
    def parse_credentials(self):
        file_credentials = os.path.expanduser("~") + "/.aws/credentials"

        if not Path(file_credentials).is_file:
            return

        profiles = {}
        with open(file_credentials,'r') as file:
            for line in file:
                if len(line.strip()) > 0:
                    profile = re.search("(\[.*\])", line) 
                    if profile:
                        profile_name = profile.group(1).replace("[","").replace("]","")
                        profiles[profile_name] = {}
    
                    attribute = re.search("(.*=.*)", line) 
                    if attribute:
                        attribute_pair  = attribute.group(1).split("=")
                        attribute_key   = attribute_pair[0].strip()
                        attribute_value = attribute_pair[1].strip()
                        profiles[profile_name][attribute_key] = attribute_value
        
        self.credentials["profiles"] = profiles
    
    def botoSession(self, profile):
        if not profile:
           return boto3.Session()
        return boto3.Session(profile_name=profile)

    def getAccountOwner(self, profile=None):
        sts = self.botoSession(profile).client("sts")
        return sts.get_caller_identity()
    
    def getMFADevices(self, profile=None):
        iam = self.botoSession(profile).client("iam")
        return iam.list_mfa_devices()
    
    def session_info(self):
        sizeSeparator    = 28
        sizeLabelProfile = 18

        try:
            sts = self.getAccountOwner(self.config.awsProfile)
        except ClientError as e:
            if e.response['Error']['Code'] == "ExpiredToken":
                msg = e.response['Error']['Message']
                Messages.showError("Expired Token!")
                return ""
        
        output      = "\n\n"
        assume_role = self.config.assumeRole if self.config.assumeRole and self.config.assumeRole != "" else "---"
        mfa_device  = self.config.mfaSerial if self.config.mfaSerial and self.config.mfaSerial != "" else "---"

        cookie = self.config.loadCookieTempCredentials()
        if not cookie:
            output    += f"{Style.GREEN}" + "=".ljust(sizeSeparator,"=") + "\n"
            output    += (f"    {Style.IYELLOW}{Emoticons.pin()} AWS SESSION") + "\n"
            output    += f"{Style.GREEN}" + "=".ljust(sizeSeparator,"=") + f"\n{Style.GREEN}"
        else:
            sizeSeparator = 44
            expiration    = cookie.timeToExpire()
            access_key    = cookie.accessKey
            output       += f"{Style.GREEN}" + "=".ljust(sizeSeparator,"=") + "\n"
            output       += (f"    {Style.IYELLOW}{Emoticons.pin()} TEMPORARY AWS SESSION (TOKEN)") + "\n"
            output       += f"{Style.GREEN}" + "=".ljust(sizeSeparator,"=") + f"\n{Style.GREEN}"

            if expiration == -1:
                expiration = f"{Style.IRED}TIME OUT {Style.GREEN}Expired!"
        
        output += f"  Profile............:{Style.IBLUE} {self.config.awsProfile} {Style.GREEN}\n"
        if cookie:
            output += f"  Access Key.........:{Style.IBLUE} {access_key}{Style.GREEN}\n"
        output += f"  Account............:{Style.IBLUE} {sts['Account']} {Style.GREEN}\n"
        output += f"  Region.............:{Style.IBLUE} {self.config.awsRegion} {Style.GREEN}\n"
        output += f"  Assume Role........:{Style.IBLUE} {assume_role} {Style.GREEN}\n"
        output += f"  MFA Device.........:{Style.IBLUE} {mfa_device} {Style.GREEN}\n"
        if cookie:
            output += f"  Remaining Time.....:{Style.IBLUE} {expiration}{Style.RESET}\n"

        return output


import os, sys, re
from os.path import expanduser
from datetime import datetime
from tinydb import Query, where
from pyawscp.messages import Messages
from pyawscp.Utils import Utils, Style
from pyawscp.Emoticons import Emoticons
from pyawscp.Config import Config
from pyawscp.PrettyTable import PrettyTable
from pyawscp.TableArgs import TableArgs
from pyawscp.Aws_Session import AwsSession
from pyawscp.Functions import Functions
from pyawscp.repository.repository_manager import CredentialsRepository, MfaRepository, RoleRepository

sizeSeparator    = 120
sizeLabelProfile = 18

class CredentialsService:

    def __init__(self,config):
        self.preferences            = config
        self.credentials_repository = CredentialsRepository()
        self.role_repository        = RoleRepository()
        self.mfa_repository         = MfaRepository()

    def loadAndSyncProfiles(self):
        self.credentials_repository.purge()

        fileCredentials  = expanduser("~") + "/.aws/credentials"
        if not os.path.exists(fileCredentials):
            Messages.showWarning("Credentials Not Found!",f" The file {Style.IBLUE}/.aws/credentials{Style.GREEN} was not found, did you install and configure your {Style.IBLUE}awscli{Style.GREEN}?")
            sys.exit()

        currentProfileName                   = None
        accessKey                            = None
        secretKey                            = None
        credentialsProfiles                  = {}
        credentialsProfiles["credentials"]   = []
        with open(fileCredentials,'r') as fcred:
            for line in fcred.readlines():
                if line.lstrip().rstrip().startswith("["):
                   if currentProfileName:
                       # SAVE Previous
                       credentialsProfiles["credentials"].append({
                           "profile": currentProfileName,
                           "accessKey": accessKey,
                           "secretKey": secretKey
                       })
                   # Read Current    
                   currentProfileName = Utils.cleanBreakLines(line).replace("[","").replace("]","") 
                else:
                   if Utils.cleanBreakLines(line).lstrip().rstrip() != "":
                      line = Utils.cleanBreakLines(re.sub('\s',"",line))
                      kv   = line.split("=")
                      if   "aws_access_key_id" in kv[0].lower():
                         accessKey = kv[1]
                      elif "aws_secret_access_key" in kv[0].lower():
                         secretKey = kv[1]
            if currentProfileName:
                # SAVE Previous
                credentialsProfiles["credentials"].append({
                    "profile": currentProfileName,
                    "accessKey": accessKey,
                    "secretKey": secretKey
                })

        # Check if needs updates/inserts
        aws_session = AwsSession(self.preferences)
        total_loaded = 0
        for c in credentialsProfiles["credentials"]:
             dbRecordProfile = self.credentials_repository.findByProfile(c["profile"])
             # Insert the new one
             if not dbRecordProfile:
                 account   = ""
                 # Read Account and MFA Devices of the Profile at AWS
                 try:
                     account = aws_session.getAccountOwner(c["profile"])["Account"]
                     for m in aws_session.getMFADevices(c["profile"])["MFADevices"]:
                         mfaRecord = self.mfaRepository.searchByQuery(Query()['mfa-device'] == m["SerialNumber"])
                         if not mfaRecord or len(mfaRecord) < 0:
                             self.mfaRepository.insert({
                                 "profile": c["profile"],
                                 "mfa-device": m["SerialNumber"],
                                 "user-name": m["UserName"]
                             })
                 except:
                     #:XXX  Check this later
                     # LogManager().LOG.warning(f"Not able to retrieve the MFA Devices for the {c['profile']}, probably some police might be blocking by the lack of MFA Token")
                     pass
                 c["account"] = account
                 self.credentials_repository.insert(c)
                 total_loaded += 1
             else:
                 # Check if needs to update the record (credentials change)
                 if dbRecordProfile["accessKey"] != c["accessKey"]  or \
                    dbRecordProfile["secretKey"] != c["secretKey"]:
                    # UPDATEIT
                    self.credentials_repository.update(c, c["profile"])

        output  = "\n\n"
        output += f"{Style.GREEN}   {Emoticons.pointRight()} Total Credentials Loaded/Synchronized...:{Style.IBLUE} {total_loaded}\n"
        output += f"{Style.GREEN}      You can see then:{Style.IBLUE} {Functions.LIST_PROFILES}"
        output += "\n"
        return output
    
    def get_profile(self, profile):
        return self.credentials_repository.findByProfile(profile)
        
    def list_profiles(self, list_roles=False, list_mfas=False):
        profiles = self.credentials_repository.all()
        if len(profiles) == 0:
            print("")
            Messages.showWarning("No AWS Credentials loaded in PyAwsCp, yet!")
            output  = f"{Style.GREEN}    You can {Style.IYELLOW}load {Style.GREEN}and {Style.IYELLOW}synchronize{Style.GREEN} in PyAwsCp your credentials from {Style.ICYAN}~/.aws/credentials\n" 
            output += f"{Style.GREEN}    For that, start using the command: {Style.BBLUE}syncCredentials" 
            output += f"{Style.RESET}" 
            print(output)
            return None
        
        if list_roles:
            for p in profiles:
                roles = self.role_repository.searchByQuery(Query().profile == p['profile'])
                if roles and len(roles) > 0:
                    p['roles'] = []
                    for idx, role in enumerate(roles):
                        p['roles'].append({
                            "name": role['role-name'],
                            "arn": role['role-arn']
                        })
        
        if list_mfas:
            for p in profiles:
                mfas = self.mfa_repository.searchByQuery(Query().profile == p['profile'])
                if mfas and len(mfas) > 0:
                    p['mfas'] = []
                    for idx, mfa in enumerate(mfas):
                        p['mfas'].append({
                            "mfa-device": mfa['mfa-device'],
                            "user-name": mfa['user-name']
                        })
            
        return profiles
    
    def list_profiles_view(self):
        list_roles = False
        list_mfas  = False
        if "roles" in self.preferences.commandArguments:
            list_roles = True
        if "mfas" in self.preferences.commandArguments:
            list_mfas = True

        profiles_loaded = self.list_profiles(list_roles,list_mfas)
        if not profiles_loaded:
            result = f"{Style.GREEN}No Profiles found" + "\n\n"
            return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_PROFILES]["name"], result, self.preferences, "", True, None)
 
        tableArgs  = TableArgs()
        tableArgs.setArguments(self.preferences.commandArguments) 
        self.preferences.awsTagsToFilter = None

        jsonResult = ""
        if self.preferences.printResults or tableArgs.verbose or tableArgs.saveToFile: 
           jsonResult = Utils.dictToJson(profiles_loaded)

        header = ["#","Profile", "Access Key", "Account"]
        if list_roles:
            header.append("Role")
        elif list_mfas:
            header.append("MFA Device")
        prettyTable = PrettyTable(header)

        idx_lin    = 0
        for p in profiles_loaded:
            idx_lin       += 1
            profile        = p['profile']
            aws_access_key = p['accessKey']
            account        = p['account']

            if list_roles:
                if 'roles' in p:
                    for idx_role, r in enumerate(p['roles']):
                        role = f"{Style.IBLUE}{idx_role+1:02d}{Style.GREEN}-{r['arn']}{Style.IBLUE} ({r['name']}){Style.GREEN}"
                        columns = [ idx_lin, profile, aws_access_key, account, role ]
                        prettyTable.addRow(columns)
                        profile = ""
                        aws_access_key = ""
                        account = ""
                else:
                    columns = [ idx_lin, profile, aws_access_key, account, "" ]
                    prettyTable.addRow(columns)
            elif list_mfas:
                if 'mfas' in p:
                    for idx_mfa, r in enumerate(p['mfas']):
                        mfa_device = f"{Style.IBLUE}{idx_mfa+1:02d}{Style.GREEN}-{r['mfa-device']}{Style.IBLUE} ({r['user-name']}){Style.GREEN}"
                        columns = [ idx_lin, profile, aws_access_key, account, mfa_device ]
                        prettyTable.addRow(columns)
                        profile = ""
                        aws_access_key = ""
                        account = ""
                else:
                    columns = [ idx_lin, profile, aws_access_key, account, "" ]
                    prettyTable.addRow(columns)
            else:
                columns = [ idx_lin, profile, aws_access_key, account ]
                prettyTable.addRow(columns)
        
        resultTxt = ""
        resultTxt = " Total...: " + Style.GREEN + format(idx_lin,",") + Style.RESET

        if (int(tableArgs.sortCol) - 1) > len(columns):
            tableArgs.sortCol = "1"
            prettyTable.sortByColumn(int(tableArgs.sortCol) - 1)
        
        tableArgs.setArguments(f"| {self.preferences.awsProfile}")

        prettyTable.sortByColumn(0)   
        prettyTable.numberSeparator = True
        prettyTable.ascendingOrder(not tableArgs.desc)
        result = resultTxt + "\n\n" + prettyTable.printMe("listProfiles",self.preferences.tableLineSeparator, tableArgs)
        return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_PROFILES]["name"], result, self.preferences, jsonResult, True, tableArgs)
    
    def import_from_awsee(self):
        print("")
        awsee_folder   = os.path.join(expanduser("~"),".awsee")
        pyawscp_folder = os.path.join(expanduser("~"),".pyawscp")

        if not os.path.exists(awsee_folder):
            Messages.showError(f"{Style.GREEN}AwsSee home folder was not found!\n    Are you sure you have it installed?\n    {Style.CYAN}https://ualter.github.io/awsee-site/")
            return
        
        snapshot_date = f"{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}"

        # CREDENTIALS
        awsee_credentials_file = os.path.join(awsee_folder,"cred")
        if os.path.exists(awsee_credentials_file):
            with open(awsee_credentials_file, 'r') as f:
                awsee_credentials_file_content = f.read() 

            # Backup PyAwsCp Credentials File
            pyawscp_credentials_file        = os.path.join(pyawscp_folder,"credentials")
            pyawscp_credentials_file_backup = os.path.join(pyawscp_folder,f"bkp-credentials-{snapshot_date}")
            if os.path.exists(pyawscp_credentials_file):
                with open(pyawscp_credentials_file, 'r') as f:
                    pyawscp_credentials_file_content = f.read()
                with open(pyawscp_credentials_file_backup,'w') as f:
                    f.write(pyawscp_credentials_file_content)
            # Import Credentials from AwsSee to PyAwsCp        
            with open(pyawscp_credentials_file,'w') as f:
                f.write(awsee_credentials_file_content)
        
        # ROLES
        awsee_roles_file = os.path.join(awsee_folder,"role")
        if os.path.exists(awsee_roles_file):
            with open(awsee_roles_file, 'r') as f:
                awsee_roles_file_content = f.read() 

            # Backup PyAwsCp Role File
            pyawscp_role_file        = os.path.join(pyawscp_folder,"role")
            pyawscp_role_file_backup = os.path.join(pyawscp_folder,f"bkp-role-{snapshot_date}")
            if os.path.exists(pyawscp_role_file):
                with open(pyawscp_role_file, 'r') as f:
                    pyawscp_role_file_content = f.read()
                with open(pyawscp_role_file_backup,'w') as f:
                    f.write(pyawscp_role_file_content)
            # Import Role from AwsSee to PyAwsCp        
            with open(pyawscp_role_file,'w') as f:
                f.write(awsee_roles_file_content)
        
        # MFA
        awsee_mfas_file = os.path.join(awsee_folder,"mfa")
        if os.path.exists(awsee_mfas_file):
            with open(awsee_mfas_file, 'r') as f:
                awsee_mfas_file_content = f.read() 

            # Backup PyAwsCp MFA File
            pyawscp_mfa_file        = os.path.join(pyawscp_folder,"mfa")
            pyawscp_mfa_file_backup = os.path.join(pyawscp_folder,f"bkp-mfa-{snapshot_date}")
            if os.path.exists(pyawscp_mfa_file):
                with open(pyawscp_mfa_file, 'r') as f:
                    pyawscp_mfa_file_content = f.read()
                with open(pyawscp_mfa_file_backup,'w') as f:
                    f.write(pyawscp_mfa_file_content)
            # Import MFA from AwsSee to PyAwsCp        
            with open(pyawscp_mfa_file,'w') as f:
                f.write(awsee_mfas_file_content)

        Messages.showMessage(f"Imported susccesfully!\n Backup created with label date {Style.CYAN}{snapshot_date}{Style.GREEN}")
    
    def rollback_import_from_awsee(self):
        #from_date = "2021-10-0610-11-13"
        print("")
        if self.preferences.commandArguments == "":
            Messages.showError(f"{Style.GREEN} Date and Time not informed!\n    Expected format: {Style.CYAN}YYYY-MM-DD_HH-MM-SS\n{Style.GREEN}    Example........: {Style.CYAN}2021-10-06_10-11-13")
            return

        from_date = self.preferences.commandArguments

        correct_format = re.match("\d{4}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}",from_date)
        if not correct_format:
            Messages.showError(f"{Style.GREEN}Wrong format!\n    Expected format: {Style.CYAN}YYYY-MM-DD_HH-MM-SS\n{Style.GREEN}    Example........: {Style.CYAN}2021-10-06_10-11-13")
            return
        
        pyawscp_folder = os.path.join(expanduser("~"),".pyawscp")

        # Restore Credentials
        pyawscp_credentials_file        = os.path.join(pyawscp_folder,"credentials")
        pyawscp_credentials_file_backup = os.path.join(pyawscp_folder,f"bkp-credentials-{from_date}")
        if not os.path.exists(pyawscp_credentials_file_backup):
            Messages.showError(f"The file bkp-credentials-{from_date} was not found!")
            return
        with open(pyawscp_credentials_file_backup, 'r') as f:
            content = f.read()
        with open(pyawscp_credentials_file,'w') as f:
            f.write(content)
        
        # Restore Roles
        pyawscp_roles_file        = os.path.join(pyawscp_folder,"role")
        pyawscp_roles_file_backup = os.path.join(pyawscp_folder,f"bkp-role-{from_date}")
        if not os.path.exists(pyawscp_roles_file_backup):
            Messages.showError(f"The file bkp-role-{from_date} was not found!")
            return
        with open(pyawscp_roles_file_backup, 'r') as f:
            content = f.read()
        with open(pyawscp_roles_file,'w') as f:
            f.write(content)
        
        # Restore MFAs
        pyawscp_mfas_file        = os.path.join(pyawscp_folder,"mfa")
        pyawscp_mfas_file_backup = os.path.join(pyawscp_folder,f"bkp-mfa-{from_date}")
        if not os.path.exists(pyawscp_mfas_file_backup):
            Messages.showError(f"The file bkp-mfa-{from_date} was not found!")
            return
        with open(pyawscp_mfas_file_backup, 'r') as f:
            content = f.read()
        with open(pyawscp_mfas_file,'w') as f:
            f.write(content)

        Messages.showMessage(f" Backup from {Style.CYAN}{from_date}{Style.GREEN} restored!")


import os, sys, re, uuid
from arnparse import arnparse
from os.path import expanduser
from tinydb import Query, where
from pyawscp.messages import Messages
from pyawscp.Utils import Utils, Style
from pyawscp.Emoticons import Emoticons
from pyawscp.Config import Config
from pyawscp.Functions import Functions
from pyawscp.PrettyTable import PrettyTable
from pyawscp.TableArgs import TableArgs
from pyawscp.Aws_Session import AwsSession
from pyawscp.repository.credentials_service import CredentialsService
from pyawscp.repository.repository_manager import MfaRepository

sizeSeparator    = 120
sizeLabelProfile = 18

class MfaService:

    def __init__(self,config):
        self.preferences         = config
        self.credentials_service = CredentialsService(config)
        self.mfa_repository     = MfaRepository()
    
    def list_mfas_by_profile(self, profile):
        return self.mfa_repository.findByProfile(profile)

    def search_mfa_by_username(self, username):
        return self.mfa_repository.searchByQuery(Query()['user-name'] == username)

    def list_mfas(self):
        return self.mfa_repository.all()

    def list_mfas_view(self):
        mfas = self.list_mfas()

        tableArgs  = TableArgs()
        tableArgs.setArguments(self.preferences.commandArguments) 
        self.preferences.awsTagsToFilter = None

        jsonResult = ""
        if self.preferences.printResults or tableArgs.verbose or tableArgs.saveToFile: 
           jsonResult = Utils.dictToJson(mfas)

        if not mfas or len(mfas) == 0:
            result = f"{Style.GREEN}No MFA Serial found" + "\n\n"
            return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_MFAS]["name"], result, self.preferences, jsonResult, True, tableArgs)

        header = ["#","MFA Device/Serial","Username","Profile","Account"]
        prettyTable = PrettyTable(header)

        idx_lin    = 0
        for m in mfas:
            idx_lin       += 1

            account = ""
            profile = self.credentials_service.get_profile(m['profile'])
            if profile:
                account = profile['account']
                
            columns = [ idx_lin, m['mfa-device'], m['user-name'], m['profile'], account]

            prettyTable.addRow(columns)
        
        resultTxt = ""
        resultTxt = " Total...: " + Style.GREEN + format(idx_lin,",") + Style.RESET

        if (int(tableArgs.sortCol) - 1) > len(columns):
            tableArgs.sortCol = "1"
            prettyTable.sortByColumn(int(tableArgs.sortCol) - 1)
        
        tableArgs.setArguments(f"| {self.preferences.awsProfile}")

        prettyTable.sortByColumn(0)   
        prettyTable.numberSeparator = True
        prettyTable.ascendingOrder(not tableArgs.desc)
        result = resultTxt + "\n\n" + prettyTable.printMe("listMfas",self.preferences.tableLineSeparator, tableArgs)
        return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_ROLES]["name"], result, self.preferences, jsonResult, True, tableArgs)

    def _choose_profile_for_mfa(self):
        output           = ""
        index            = 0
        profiles         = []

        output += (f" {Style.BBLUE}   Which AWS Profile Credentials?") + "\n"
        output += (f"{Style.GREEN}" + "=".ljust(sizeSeparator,"=")) + "\n"
        
        for profile in self.credentials_service.list_profiles():
            index += 1
            profiles.append(profile)
            pn        = profile['profile'] + " ".rjust(sizeLabelProfile -len(profile['profile']), " ")
            account   = profile['account']
            line      = f"  {Style.IBLUE}{index:02d} -->{Style.GREEN} {pn} {Style.IBLUE}Account..:{Style.GREEN} {account}"
            labelMfa = "MFAs: "
            mfas     = ""
            for idx, r in enumerate(self.mfa_repository.searchByQuery(where('profile') == profile['profile'])):
                 if len(mfas) > 1:
                     mfas += "\n" + " ".ljust(len(Utils.removeCharsColors(line)) + len(labelMfa) + 1," ")
                 mfas += f"{Style.BLUE}{idx+1:02d}{Style.GREEN}-{r['mfa-device']} {Style.BLUE}({r['user-name']})"
            if len(mfas) > 1:
                mfas = f"{Style.IBLUE}{labelMfa}{Style.GREEN}{mfas}{Style.GREEN}"
            output += line + f" {mfas}\n"
        output += "-".ljust(sizeSeparator,"-")
        print(output)
        print(f"{Style.GREEN} {Emoticons.pin()} Choose a profile....:{Style.BLUE}", end= ' ')
        result = input()
        if result == "":
            return None
            
        if not Utils.isNumber(result):
            Messages.showWarning(f"Invalid choice \"{Style.IBLUE}{result}{Style.IGREEN}\"")
            return None
        if int(result) > index:
            Messages.showWarning(f"Profile \"{Style.IBLUE}{result}{Style.IGREEN}\" not valid!")
            return None
        
        return profiles[int(result) - 1]

    def add_mfa(self):
        output  = ""
        output += f"{Style.GREEN}" + "=".ljust(sizeSeparator,"=") + "\n"
        output += (f" {Style.BYELLOW}{Emoticons.pin()} ADD A MFA SERIAL")
        print(output)

        profile_chosen = self._choose_profile_for_mfa()
        if not profile_chosen:
            print("\n")
            return 
        
        mfa       = {}
        print(f"{Style.GREEN} {Emoticons.pin()} Device..............:{Style.BLUE}",end=' ')
        mfaDevice = input()
        spacesFound = re.search(" ",mfaDevice)
        if spacesFound:
            Messages.showWarning(f"Invalid Device \"{Style.IBLUE}{mfaDevice}{Style.IGREEN}\", spaces are not allowed")
            print("\n")
            return None
        try:
            mfaArnParse = arnparse(mfaDevice)
        except:
            Messages.showWarning(f"Invalid MFA Serial Device ARN \"{Style.IBLUE}{mfaDevice}{Style.IGREEN}\"")
            print("\n")
            return None
        mfa["mfa-device"] = mfaDevice

        print(f"{Style.GREEN} {Emoticons.pin()} Username............:{Style.BLUE}",end=' ')
        mfaUserName = input()
        if mfaUserName == "":
            Messages.showWarning(f"Invalid Username \"{Style.IBLUE}{mfaUserName}{Style.IGREEN}\"")
            print("\n")
            return None
        mfa["user-name"] = mfaUserName

        output  = "\n"
        output += f" {Style.IBLUE} --> {Style.GREEN}Profile............: {Style.IBLUE}{profile_chosen['profile']}{Style.RESET}\n"
        output += f" {Style.IBLUE} --> {Style.GREEN}Account............: {Style.IBLUE}{profile_chosen['account']}{Style.RESET}\n"
        output += f" {Style.IBLUE} --> {Style.GREEN}Device.............: {Style.IBLUE}{mfaDevice}{Style.RESET}\n"
        output += f" {Style.IBLUE} --> {Style.GREEN}Username...........: {Style.IBLUE}{mfaUserName}{Style.RESET}\n"
        print(output)

        print(f"{Style.GREEN} {Emoticons.pin()} Confirm [{Style.IGREEN}y/{Style.IGREEN}n{Style.RESET}]:{Style.BLUE}",end=' ')
        confirm = input()
        if confirm.lower() == "y":
            record = {
                "id": str(uuid.uuid4()),
                "profile": profile_chosen['profile'],
                "mfa-device": mfaDevice,
                "user-name": mfaUserName
            }
            self.mfa_repository.insert(record)
            output = f"{Style.GREEN} {Emoticons.ok()} Saved!{Style.RESET}"
            print(output)
    
    def remove_mfa(self):
        output  = ""
        output += f"{Style.GREEN}" + "=".ljust(sizeSeparator,"=") + "\n"
        output += (f" {Style.BYELLOW}{Emoticons.pin()} REMOVE A ROLE\n")
        output += (f" {Style.BBLUE}   Which Mfa?") + "\n"
        output += (f"{Style.GREEN}" + "=".ljust(sizeSeparator,"=")) + "\n"

        mfas = []
        index = 0
        for r in self.list_mfas():
            index += 1
            mfas.append(r)
            pn        = r['mfa-device']  + " ".rjust(sizeLabelProfile -len(r['mfa-device']), " ")
            name      = r['user-name'] + " ".rjust(20 - len(r['user-name']), " ")
            line      = f"  {Style.IBLUE}{index:02d} -->{Style.GREEN} {pn} {Style.IBLUE}{name}{Style.GREEN} (Profile:{Style.CYAN} {r['profile']}{Style.GREEN})"
            output += line + "\n"
        output += "-".ljust(sizeSeparator,"-")
        print(output)
        print(f"{Style.GREEN} {Emoticons.pin()} Choose a mfa....:{Style.BLUE}", end= ' ')
        result = input()
        if result == "":
            print("\n")
            return None
        if not Utils.isNumber(result) or (int(result) > len(mfas)) or (int(result) == 0):
            Messages.showWarning(f"Invalid choice \"{Style.IBLUE}{result}{Style.IGREEN}\"")
            print("\n")
            return None

        numberInList = int(result)
        mfaToRemove = None
        
        for idx,mfa in enumerate(mfas):
            if (idx+1) == numberInList:
                mfaToRemove = mfa
                break
        
        if not mfaToRemove:
            Messages.showWarning("Mfa not found!")
            print("\n")
            return
        else:
            output  = "\n"
            output += f" {Style.IBLUE} --> {Style.GREEN}MFA Device.........: {Style.IBLUE}{mfaToRemove['mfa-device']}{Style.RESET}\n"
            output += f" {Style.IBLUE} --> {Style.GREEN}Username...........: {Style.IBLUE}{mfaToRemove['user-name']}{Style.RESET}\n"
            output += f" {Style.IBLUE} --> {Style.GREEN}Profile............: {Style.IBLUE}{mfaToRemove['profile']}{Style.RESET}\n"
            print(output)
            print(f"{Style.GREEN} {Emoticons.pin()} Remove [{Style.IGREEN}y/{Style.IGREEN}n{Style.RESET}]:{Style.BLUE}",end=' ')
            confirm = input()
            if confirm.lower() == "y":
                self.mfa_repository.remove(Query().id == mfaToRemove['id'])
                output  = "\n"
                output += f"{Style.GREEN} {Emoticons.ok()} Removed!{Style.RESET}"
                print(output)




# coding=utf-8

import sys
import json
import os, re
import random
import configparser
from os.path import expanduser

from botocore.credentials import Credentials
from pyawscp.Functions import Functions
from pyawscp.PyAwsShell import PREFERENCES
from pyawscp.Utils import Utils, Style
from pyawscp.Emoticons import Emoticons
from pyawscp.Config import Config
from pyawscp.TableArgs import TableArgs
from pygments import highlight, lexers, formatters
from pyawscp.PyEc2Cp import PyEc2CP
from pyawscp.PyEc2NetSecCp import PyEc2NetSecCP
from pyawscp.PyElbCp import PyElbCP
from pyawscp.PyS3Cp import PyS3CP
from pyawscp.PyR53Cp import PyR53CP
from tinydb import Query, where
from datetime import datetime
from arnparse import arnparse

from pyawscp.Aws_Session import AwsSession as AwsSession
from pyawscp.messages import Messages
from pyawscp.repository import credentials_service, mfa_service
from pyawscp.repository import role_service
from repository.repository_manager import RoleRepository
from repository.credentials_service import CredentialsService
from repository.role_service import RoleService
from repository.mfa_service import MfaService

class Test:
    def __init__(self):
        self.my_function(a=12, _vpcid=123456)

    def my_function(self, **args):
        print(args["_vpcid"])

def loadTmpSession():
    tempConfigFile = os.path.expanduser("~") + "/.pyawscp/tmp.bin"
    if not os.path.exists(tempConfigFile):
       print("")
       print("=".ljust(80,"="))
       print("File {} not found".format(tempConfigFile))
       print("""
       FILE NAME: \033[32mmyTempSession.tmp\033[0m
       EXPECTED FIELDS:\033[32m
       assume-role=NaNaNaNa
       mfa-serial=NaNaNaNa
       region=NaNaNaNa
       profile=NaNaNaNa\033[0m
       """)
       print("=".ljust(80,"="))
       print("")
       sys.exit()
    
    config = Config()
    with open(tempConfigFile, "r") as fp:
        for idx, line in enumerate(fp):
            if not line.startswith("#"):
                if "assume-role=" in line:
                    config.assumeRole = line.replace("assume-role=","").rstrip().lstrip()
                elif "mfa-serial=" in line:
                    config.mfaSerial = line.replace("mfa-serial=","").rstrip().lstrip() 
                elif "region=" in line:
                    config.awsRegion = line.replace("region=","").rstrip().lstrip() 
    config.printResults = False
    config.tableLineSeparator = False

    if len(sys.argv) > 1:
       config.mfaPassCode=sys.argv[1]
    
    print("=".ljust(80,"="))
    print(" All set!")
    print("=".ljust(80,"="))
    return config

def loadSavedConfig():
    preferences = Config()
    configFileIni = configparser.ConfigParser()
    configFileIni.read(expanduser("~") + "/.pyawscp/pyawscp.ini")
    preferences.awsProfile               = configFileIni[PREFERENCES]["aws-profile"]
    preferences.awsRegion                = configFileIni[PREFERENCES]["aws-region"]
    preferences.assumeRole               = configFileIni[PREFERENCES]["assume-role"]
    preferences.mfaSerial                = configFileIni[PREFERENCES]["mfa-serial"]
    return preferences

def test2():
    config = loadTmpSession()
    pyEc2CP = PyEc2CP(config)
    connBt, data, report, content = pyEc2CP.listEc2()
    print(report)

def testListVpc():
    config = loadTmpSession()
    config.commandArguments = "verbose"
    pyEc2Cp = PyEc2CP(config)
    report = pyEc2Cp.listVpc()
    print(report)

def testListEc2():
    config = loadTmpSession()
    pyEc2Cp = PyEc2CP(config)
    connBt, data, report, content = pyEc2Cp.listEc2()
    print(report)

def testFindRoute():
    config = loadTmpSession()
    #config.commandArguments = "i-0b5d342791d7ed2c4,i-0c54d18dc352d71e6,excel"
    #config.commandArguments = "i-0b5d342791d7ed2c4,i-0c54d18dc352d71e6,save,verbose"
    #config.commandArguments = "i-0b5d342791d7ed2c4,i-0c54d18dc352d71e6,target-group"
    #config.commandArguments = "i-0b5d342791d7ed2c4,i-0c54d18dc352d71e6"
    #config.commandArguments = "i-06f4a5c3bb81fec46,i-0d9a673532481f016,i-09117ce6865a30ab7,i-04a483aed2cf4a8c7"
    config.commandArguments = "i-06f4a5c3bb81fec46,i-0d9a673532481f016,i-09117ce6865a30ab7"

    #config.commandArguments = "i-0d9a673532481f016,i-09117ce6865a30ab7,label-name,target-group"
    #config.commandArguments = "i-0d9a673532481f016,i-09117ce6865a30ab7,target-group"
    #config.commandArguments = "i-0d9a673532481f016,i-09117ce6865a30ab7,label-name"
    config.tableLineSeparator = False

    Utils.clearScreen()
    py = PyR53CP(config)
    data, report = py.findRouteEc2s()
    print(Utils.formatPrettyJson(data))
    #print(report)

    # Utils.saveToFile("TestData.json",data)
    # Load previous data (saved before) to save time on testing only Result Format
    # with open("results/TestData.json", 'r') as f:
    #     jsonSaved = f.read()
    # data = json.loads(jsonSaved)    

    result = py.findRouteEc2sPrintResult(data, config.commandArguments)
    print(result)

def test_list_profiles():
    config = loadTmpSession()
    credentials_service = CredentialsService(config)   
    print(credentials_service.list_profiles_view())

def test_list_roles():
    config = loadTmpSession()
    role_service = RoleService(config)
    print(role_service.list_roles_view())

def test_feature_sync_repository_credentials():
    config = loadTmpSession()

    print("")
    Messages.showStartExecution("Wait, synchronizing AWS credentials...      ")
    credentials_service = CredentialsService(config)
    credentials_service.loadAndSyncProfiles()
    Messages.showStartExecution("Synchronization done!                        ")
    print("")

def test_feature_list_repository_credentials():
    config = loadTmpSession()
    print("")
    credentials_service = CredentialsService(config)
    print(credentials_service.list_profiles_view())

def test_feature_add_role():
    config = loadTmpSession()
    role_service = RoleService(config)
    role_service.add_role()

def test_remove_role():
    config = loadTmpSession()
    role_service = RoleService(config)
    role_service.remove_role()

def test_list_mfas():
    config = loadTmpSession()
    mfa_service = MfaService(config)
    print(mfa_service.list_mfas_view())

def test_add_mfa():
    config = loadTmpSession()
    mfa_service = MfaService(config)
    mfa_service.add_mfa()

def test_remove_mfa():
    config = loadTmpSession()
    mfa_service = MfaService(config)
    mfa_service.remove_mfa()

def rollback_import_from_awsee():
    from_date = "2021-10-06_13-06-55"
    #from_date = "2021-10-0610-11-13"
    correct_format = re.match("\d{4}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}",from_date)
    if not correct_format:
        Messages.showError(f"{Style.GREEN}Wrong format!\n    Expected format: {Style.CYAN}YYYY-MM-DD_HH-MM-SS\n{Style.GREEN}    Example........: {Style.CYAN}2021-10-06_10-11-13")
        return
    
    pyawscp_folder = os.path.join(expanduser("~"),".pyawscp")

    # Restore Credentials
    pyawscp_credentials_file        = os.path.join(pyawscp_folder,"credentials")
    pyawscp_credentials_file_backup = os.path.join(pyawscp_folder,f"bkp-credentials-{from_date}")
    if not os.path.exists(pyawscp_credentials_file_backup):
        Messages.showError(f"The file bkp-credentials-{from_date} was not found!")
        return
    with open(pyawscp_credentials_file_backup, 'r') as f:
        content = f.read()
    with open(pyawscp_credentials_file,'w') as f:
        f.write(content)
    
    # Restore Roles
    pyawscp_roles_file        = os.path.join(pyawscp_folder,"role")
    pyawscp_roles_file_backup = os.path.join(pyawscp_folder,f"bkp-role-{from_date}")
    if not os.path.exists(pyawscp_roles_file_backup):
        Messages.showError(f"The file bkp-role-{from_date} was not found!")
        return
    with open(pyawscp_roles_file_backup, 'r') as f:
        content = f.read()
    with open(pyawscp_roles_file,'w') as f:
        f.write(content)
    
    # Restore MFAs
    pyawscp_mfas_file        = os.path.join(pyawscp_folder,"mfa")
    pyawscp_mfas_file_backup = os.path.join(pyawscp_folder,f"bkp-mfa-{from_date}")
    if not os.path.exists(pyawscp_mfas_file_backup):
        Messages.showError(f"The file bkp-mfa-{from_date} was not found!")
        return
    with open(pyawscp_mfas_file_backup, 'r') as f:
        content = f.read()
    with open(pyawscp_mfas_file,'w') as f:
        f.write(content)

    Messages.showMessage(f" Backup from {Style.CYAN}{from_date}{Style.GREEN} restored!")

def import_from_awsee():
    awsee_folder   = os.path.join(expanduser("~"),".awsee")
    pyawscp_folder = os.path.join(expanduser("~"),".pyawscp")

    if not os.path.exists(awsee_folder):
        Messages.showError(f"{Style.GREEN}AwsSee home folder was not found!\n    Are you sure you have it installed?\n    {Style.CYAN}https://ualter.github.io/awsee-site/")
        return
    
    snapshot_date = f"{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}"

    # CREDENTIALS
    awsee_credentials_file = os.path.join(awsee_folder,"cred")
    if os.path.exists(awsee_credentials_file):
        with open(awsee_credentials_file, 'r') as f:
            awsee_credentials_file_content = f.read() 

        # Backup PyAwsCp Credentials File
        pyawscp_credentials_file        = os.path.join(pyawscp_folder,"credentials")
        pyawscp_credentials_file_backup = os.path.join(pyawscp_folder,f"bkp-credentials-{snapshot_date}")
        if os.path.exists(pyawscp_credentials_file):
            with open(pyawscp_credentials_file, 'r') as f:
                pyawscp_credentials_file_content = f.read()
            with open(pyawscp_credentials_file_backup,'w') as f:
                f.write(pyawscp_credentials_file_content)
        # Import Credentials from AwsSee to PyAwsCp        
        with open(pyawscp_credentials_file,'w') as f:
            f.write(awsee_credentials_file_content)
    
    # ROLES
    awsee_roles_file = os.path.join(awsee_folder,"role")
    if os.path.exists(awsee_roles_file):
        with open(awsee_roles_file, 'r') as f:
            awsee_roles_file_content = f.read() 

        # Backup PyAwsCp Role File
        pyawscp_role_file        = os.path.join(pyawscp_folder,"role")
        pyawscp_role_file_backup = os.path.join(pyawscp_folder,f"bkp-role-{snapshot_date}")
        if os.path.exists(pyawscp_role_file):
            with open(pyawscp_role_file, 'r') as f:
                pyawscp_role_file_content = f.read()
            with open(pyawscp_role_file_backup,'w') as f:
                f.write(pyawscp_role_file_content)
        # Import Role from AwsSee to PyAwsCp        
        with open(pyawscp_role_file,'w') as f:
            f.write(awsee_roles_file_content)
    
    # MFA
    awsee_mfas_file = os.path.join(awsee_folder,"mfa")
    if os.path.exists(awsee_mfas_file):
        with open(awsee_mfas_file, 'r') as f:
            awsee_mfas_file_content = f.read() 

        # Backup PyAwsCp MFA File
        pyawscp_mfa_file        = os.path.join(pyawscp_folder,"mfa")
        pyawscp_mfa_file_backup = os.path.join(pyawscp_folder,f"bkp-mfa-{snapshot_date}")
        if os.path.exists(pyawscp_mfa_file):
            with open(pyawscp_mfa_file, 'r') as f:
                pyawscp_mfa_file_content = f.read()
            with open(pyawscp_mfa_file_backup,'w') as f:
                f.write(pyawscp_mfa_file_content)
        # Import MFA from AwsSee to PyAwsCp        
        with open(pyawscp_mfa_file,'w') as f:
            f.write(awsee_mfas_file_content)
    
    Messages.showMessage(f"Imported susccesfully!\n Backup created with label date {Style.CYAN}{snapshot_date}{Style.GREEN}")
           

## Troubleshoot
## in WSL (Bash), in case the Enter is resulting a ^M   use this command "stty sane" before call Python
def test1():
    pass
    # val = input("Enter your value: ") 
    # print(val) 
    # AWS_CREDENTIALS_DIR = "~/.aws/"
    # print("") 
    # print("")
    # print("\033[31m ---> Ops!\033[33m AWS CREDENTIALS NOT FOUND!")
    # print("") 
    # print("") 
    # print("\033[34m ---> \033[33mPlease, configure your AWS Credentials:")
    # print("\033[34m      \033[33m1. Create the folder \033[35m{}\033[0m".format(AWS_CREDENTIALS_DIR))
    # print("\033[34m      \033[33m2. Create the file \033[35m{}credentials\033[33m with the  content:".format(AWS_CREDENTIALS_DIR))
    # print("\033[34m      \033[94m   [default]")
    # print("\033[34m      \033[94m   aws_access_key_id = YOUR_ACCESS_KEY")
    # print("\033[34m      \033[94m   aws_secret_access_key = YOUR_SECRET_KEY")
    # print("\033[34m      \033[33m3. Optionally, create the file \033[35m{}config\033[33m with your default region:".format(AWS_CREDENTIALS_DIR))
    # print("\033[34m      \033[94m   [default]")
    # print("\033[34m      \033[94m   region=us-east-1")
    # print("")
    # print("")

def session_info():
    preferences = loadSavedConfig()
    aws_session = AwsSession(preferences)
    print(aws_session.session_info())

def main():
    #test_feature_add_role()
    #test_list_profiles()
    #test_list_roles()
    #test_remove_role()
    #test_list_mfas()
    #test_add_mfa()
    #test_remove_mfa()
    #import_from_awsee()
    #rollback_import_from_awsee()
    session_info()
    

if __name__ == '__main__':
    main()
        


import os, sys, re, uuid
from arnparse import arnparse
from os.path import expanduser
from tinydb import Query, where
from pyawscp.messages import Messages
from pyawscp.Utils import Utils, Style
from pyawscp.Emoticons import Emoticons
from pyawscp.Config import Config
from pyawscp.Functions import Functions
from pyawscp.PrettyTable import PrettyTable
from pyawscp.TableArgs import TableArgs
from pyawscp.Aws_Session import AwsSession
from pyawscp.repository.credentials_service import CredentialsService
from pyawscp.repository.repository_manager import RoleRepository

sizeSeparator    = 120
sizeLabelProfile = 18

class RoleService:

    def __init__(self,config):
        self.preferences         = config
        self.credentials_service = CredentialsService(config)
        self.role_repository     = RoleRepository()
    
    def search_role_by_name(self, name):
        return self.role_repository.searchByQuery(Query()['role-name'] == name)

    def list_roles_by_profile(self, profile):
        return self.role_repository.findByProfile(profile)

    def list_roles(self):
        return self.role_repository.all()

    def list_roles_view(self):
        roles = self.list_roles()

        if not roles or len(roles) == 0:
            result = f"{Style.GREEN}No Roles found" + "\n\n"
            return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_ROLES]["name"], result, self.preferences, "", True, None)

        tableArgs  = TableArgs()
        tableArgs.setArguments(self.preferences.commandArguments) 
        self.preferences.awsTagsToFilter = None

        jsonResult = ""
        if self.preferences.printResults or tableArgs.verbose or tableArgs.saveToFile: 
           jsonResult = Utils.dictToJson(roles)

        header = ["#","Role Name","ARN","Profile","Account"]
        prettyTable = PrettyTable(header)

        idx_lin    = 0
        for r in roles:
            idx_lin       += 1

            account = ""
            profile = self.credentials_service.get_profile(r['profile'])
            if profile:
                account = profile['account']
                
            columns = [ idx_lin, r['role-name'], r['role-arn'], r['profile'], account]

            prettyTable.addRow(columns)
        
        resultTxt = ""
        resultTxt = " Total...: " + Style.GREEN + format(idx_lin,",") + Style.RESET

        if (int(tableArgs.sortCol) - 1) > len(columns):
            tableArgs.sortCol = "1"
            prettyTable.sortByColumn(int(tableArgs.sortCol) - 1)
        
        tableArgs.setArguments(f"| {self.preferences.awsProfile}")

        prettyTable.sortByColumn(0)   
        prettyTable.numberSeparator = True
        prettyTable.ascendingOrder(not tableArgs.desc)
        result = resultTxt + "\n\n" + prettyTable.printMe("listRoles",self.preferences.tableLineSeparator, tableArgs)
        return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_ROLES]["name"], result, self.preferences, jsonResult, True, tableArgs)

    def _choose_profile_for_role(self):
        output           = ""
        index            = 0
        profiles         = []

        output += (f" {Style.BBLUE}   Which AWS Profile Credentials?") + "\n"
        output += (f"{Style.GREEN}" + "=".ljust(sizeSeparator,"=")) + "\n"
        
        for profile in self.credentials_service.list_profiles():
            index += 1
            profiles.append(profile)
            pn        = profile['profile'] + " ".rjust(sizeLabelProfile -len(profile['profile']), " ")
            account   = profile['account']
            line      = f"  {Style.IBLUE}{index:02d} -->{Style.GREEN} {pn} {Style.IBLUE}Account..:{Style.GREEN} {account}"
            labelRole = "Roles..: "
            roles     = ""
            for idx, r in enumerate(self.role_repository.searchByQuery(where('profile') == profile['profile'])):
                 if len(roles) > 1:
                     roles += "\n" + " ".ljust(len(Utils.removeCharsColors(line)) + len(labelRole) + 1," ")
                 roles += f"{Style.BLUE}{idx+1:02d}{Style.GREEN}-{r['role-arn']} {Style.BLUE}({r['role-name']})"
            if len(roles) > 1:
                roles = f"{Style.IBLUE}{labelRole}{Style.GREEN}{roles}{Style.GREEN}"
            output += line + f" {roles}\n"
        output += "-".ljust(sizeSeparator,"-")
        print(output)
        print(f"{Style.GREEN} {Emoticons.pin()} Choose a profile....:{Style.BLUE}", end= ' ')
        result = input()
        if result == "":
            return None
            
        if not Utils.isNumber(result):
            Messages.showWarning(f"Invalid choice \"{Style.IBLUE}{result}{Style.IGREEN}\"")
            return None
        if int(result) > index:
            Messages.showWarning(f"Profile \"{Style.IBLUE}{result}{Style.IGREEN}\" not valid!")
            return None
        
        return profiles[int(result) - 1]

    def add_role(self):
        output  = ""
        output += f"{Style.GREEN}" + "=".ljust(sizeSeparator,"=") + "\n"
        output += (f" {Style.BYELLOW}{Emoticons.pin()} ADD A ROLE")
        print(output)

        profile_chosen = self._choose_profile_for_role()
        if not profile_chosen:
            print("\n")
            return 
        
        role       = {}
        print(f"{Style.GREEN} {Emoticons.pin()} Role Name...........:{Style.BLUE}",end=' ')
        roleName = input()
        spacesFound = re.search(" ",roleName)
        if spacesFound:
            Messages.showWarning(f"Invalid Role Name \"{Style.IBLUE}{roleName}{Style.IGREEN}\", spaces are not allowed")
            print("\n")
            return None
        role["name"] = roleName

        print(f"{Style.GREEN} {Emoticons.pin()} Role ARN............:{Style.BLUE}",end=' ')
        roleArn = input()
        try:
            roleArnParse = arnparse(roleArn)
        except:
            Messages.showWarning(f"Invalid Role ARN \"{Style.IBLUE}{roleArn}{Style.IGREEN}\"")
            print("\n")
            return None
        role["arn"] = roleArn

        output  = "\n"
        output += f" {Style.IBLUE} --> {Style.GREEN}Profile............: {Style.IBLUE}{profile_chosen['profile']}{Style.RESET}\n"
        output += f" {Style.IBLUE} --> {Style.GREEN}Account............: {Style.IBLUE}{profile_chosen['account']}{Style.RESET}\n"
        output += f" {Style.IBLUE} --> {Style.GREEN}Role Name..........: {Style.IBLUE}{roleName}{Style.RESET}\n"
        output += f" {Style.IBLUE} --> {Style.GREEN}Role ARN...........: {Style.IBLUE}{roleArn}{Style.RESET}\n"
        print(output)

        print(f"{Style.GREEN} {Emoticons.pin()} Confirm [{Style.IGREEN}y/{Style.IGREEN}n{Style.RESET}]:{Style.BLUE}",end=' ')
        confirm = input()
        if confirm.lower() == "y":
            record = {
                "id": str(uuid.uuid4()),
                "profile": profile_chosen['profile'],
                "role-name": roleName,
                "role-arn": roleArn
            }
            self.role_repository.insert(record)
            output = f"{Style.GREEN} {Emoticons.ok()} Saved!{Style.RESET}"
            print(output)
    
    def remove_role(self):
        output  = ""
        output += f"{Style.GREEN}" + "=".ljust(sizeSeparator,"=") + "\n"
        output += (f" {Style.BYELLOW}{Emoticons.pin()} REMOVE A ROLE\n")
        output += (f" {Style.BBLUE}   Which Role?") + "\n"
        output += (f"{Style.GREEN}" + "=".ljust(sizeSeparator,"=")) + "\n"
        print(output)
        roles = []
        index = 0
        for r in self.list_roles():
            index += 1
            roles.append(r)
            pn        = r['role-arn']  + " ".rjust(sizeLabelProfile -len(r['role-arn']), " ")
            name      = r['role-name'] + " ".rjust(20 - len(r['role-name']), " ")
            line      = f"  {Style.IBLUE}{index:02d} -->{Style.GREEN} {pn} {Style.IBLUE}{name}{Style.GREEN} (Profile:{Style.CYAN} {r['profile']}{Style.GREEN})"
            output += line + "\n"
        output += "-".ljust(sizeSeparator,"-")
        print(output)
        print(f"{Style.GREEN} {Emoticons.pin()} Choose a role....:{Style.BLUE}", end= ' ')
        result = input()
        if result == "":
            print("\n")
            return None
        if not Utils.isNumber(result) or (int(result) > len(roles)) or (int(result) == 0):
            Messages.showWarning(f"Invalid choice \"{Style.IBLUE}{result}{Style.IGREEN}\"")
            print("\n")
            return None

        numberInList = int(result)
        roleToRemove = None
        
        for idx,role in enumerate(roles):
            if (idx+1) == numberInList:
                roleToRemove = role
                break
        
        if not roleToRemove:
            Messages.showWarning("Role not found!")
            print("\n")
            return
        else:
            output  = "\n"
            output += f" {Style.IBLUE} --> {Style.GREEN}Role ARN...........: {Style.IBLUE}{roleToRemove['role-arn']}{Style.RESET}\n"
            output += f" {Style.IBLUE} --> {Style.GREEN}Role Name..........: {Style.IBLUE}{roleToRemove['role-name']}{Style.RESET}\n"
            output += f" {Style.IBLUE} --> {Style.GREEN}Profile............: {Style.IBLUE}{roleToRemove['profile']}{Style.RESET}\n"
            print(output)
            print(f"{Style.GREEN} {Emoticons.pin()} Remove [{Style.IGREEN}y/{Style.IGREEN}n{Style.RESET}]:{Style.BLUE}",end=' ')
            confirm = input()
            if confirm.lower() == "y":
                self.role_repository.remove(Query().id == roleToRemove['id'])
                output  = "\n"
                output += f"{Style.GREEN} {Emoticons.ok()} Removed!{Style.RESET}"
                print(output)




from genericpath import isfile
import webbrowser
import os, sys
import threading
import pathlib
import logging
import shutil
from platform import uname

from pyawscp.pymxgraph.HttpServer import HttpServer
sys.path.append('..')
from pyawscp.Utils import Utils, Style
from pyawscp.Emoticons import Emoticons

LOG = logging.getLogger("app." + __name__)
LOG.setLevel(logging.INFO)

class PymxGraphAwsNavigator:

    def __init__(self, pathToResources, images, htmlSnippets, verbose=False):
        self.pathToResources = pathToResources
        self.images          = images
        self.htmlSnippets    = htmlSnippets
        self.verbose         = verbose
        self.httpServer      = None

    def startHttpServer(self):
        self.httpServer = HttpServer(self.verbose)
        self.httpServer.run()

    def stopHttpServer(self):
        self.httpServer.stop()

    def openHTML(self,html):
        directory = os.path.dirname(os.path.realpath(__file__))
        fileName  = os.path.join(str(directory),"aws-navigator.html")
        
        try:
            os.remove(fileName)
        except OSError:
            pass

        with open(fileName, "w") as file:
            file.write(html)
        
        if os.name == "nt":
           # Windows
           # Use the user's default browser
           webbrowser.open('file://' + os.path.realpath(fileName))
        else:
           # Linux
           # Am I a WSL?  
           if 'MICROSOFT' in uname().release.upper():
               # WSL - A linux running inside Windows

               # Identifying WSL Version
               wsl_path_to_win_browser = "file://wsl%24/Ubuntu-18.04/"
               wsl_version             = None
               if os.path.isfile('/etc/os-release'):
                   with open('/etc/os-release','r') as f:
                       for line in f.readlines():
                           if "VERSION_ID" in line.upper():
                               wsl_version = line.split("=")[1].replace("\"","")
               if wsl_version:
                  wsl_path_to_win_browser = f"file://wsl%24/Ubuntu-{wsl_version}"
               
               # Check browser to use (from Windows)
               chrome  = '/mnt/c/Program Files (x86)/Google/Chrome/Application/chrome.exe'
               firefox = '/mnt/c/Program Files/Mozilla Firefox/firefox.exe'
               edge    = '/mnt/c/Program Files (x86)/Microsoft/Edge/Application/msedge.exe'
               browser = None
               if os.path.isfile(f'{chrome}'):
                  webbrowser.register('chrome',None,webbrowser.BackgroundBrowser(chrome))    
                  browser = "chrome"
               elif os.path.isfile(f'{edge}'):
                  webbrowser.register('edge',None,webbrowser.BackgroundBrowser(edge))    
                  browser = "edge" 
               elif os.path.isfile(f'{firefox}'):
                  webbrowser.register('firefox',None,webbrowser.BackgroundBrowser(firefox))    
                  browser = "firefox"
               if not browser:
                  print("")
                  print(Style.GREEN + "==================================================")
                  print(Emoticons.error() + Style.BYELLOW + " Sorry! " + Style.BLUE)
                  print("  No browser found on this system to open it!" + Style.GREEN)
                  print("==================================================")
                  print(Style.RESET)
                  return
               webbrowser.get(browser).open(f"{wsl_path_to_win_browser}{fileName}")
           else:
               # Native Linux
               webbrowser.open('file://' + os.path.realpath(fileName))

    def drawAwsNavigator(self):
        html = """
        <html>
          <head>
             """ + self.htmlSnippets.htmlTitle('Hi There!') + """
             """ + self.htmlSnippets.jsLibraries()          + """
             <style type="text/css">
                .modalInfo {
                    position: absolute;
                    display: block;
                    top: 200px;
                    left: 200px;
                    background: rgb(245, 245, 245);
                    //width: 350px;
                    //height: 250px;
                    padding: 5px;
                    font-size: 0.875rem;
                    box-shadow: 0 0 2.5rem rgba(0, 0, 0, 0.5);
                    border-radius: 0.3125rem;
                    font-family: 'Source Sans Pro', sans-serif
                }
                .content {
                    font-size: 0.8rem;
                    font-family: 'Consolas', sans-serif
                }
                .tag {
                    border: 1.4px solid darkcyan;
                    padding: 3px;
                    border-radius: 8px;
                    background-color:rgb(0,220,220);
                    color:black;
                    display:inline;
                    font-size: 12px;
                }
                .tableInfo {
                    font-family: Consolas;
                    font-size: 13px;
                    border: 2px solid darkcyan;
                    padding: 0;
                    border-radius: 8px;
                }
                .cellHeader {
                    border-bottom: 1.5px darkcyan solid;
                    font-size: 15px;
                    font-weight: bold;
                    background-color: darkcyan;
                }
                .cellInfo {
                    padding: 10px 8px 10px 8px;
                }
                .cellTags {
                    border-top: 1.5px darkcyan solid;
                    padding: 8px 8px 15px 8px;
                }
                .buttonCopyClipboard {
                    margin-bottom: -3px;
                }
             </style>
             <script type="text/javascript">

                 """ + self.htmlSnippets.jsDataURIForImage()    + """
                 
                 var requestId   = 0;

                 const Http             = new XMLHttpRequest();
                 const url              = '"""+ self.htmlSnippets.serverURL +"""';
                 var   previousGraphXml = [];
                 var   previousCoord    = {"x":0, "y":0};

                 function moveMe(graph, cellId, valueX, valueY, addToIt) {
                    if ( graph.view.getScale() == 1 ) { 
                        try
                        {
                            var v1 = graph.getModel().getCell(cellId); 
                            graph.clearSelection();
                            graph.getModel().beginUpdate(); 
                            var geo = graph.getCellGeometry(v1);
                            geo  = geo.clone();

                            previousCoord.x = geo.x;
                            previousCoord.y = geo.y;
                            
                            if ( addToIt ) {
                                geo.x += valueX;
                                geo.y += valueY;
                            } else {
                                geo.x = valueX;
                                geo.y = valueY;
                            }
                            graph.getModel().setGeometry(v1, geo);
                        }
                        finally
                        {
                            var morph = new mxMorphing(graph, 0, 0, 0);
                            morph.addListener(mxEvent.DONE, function()
                            {
                                graph.getModel().endUpdate();
                            });
                            morph.startAnimation();
                            return previousCoord;
                        }
                    }
                 }

                 function historyGraphGoBack(graph) {
                    // Clean any ModalInfo Timer started before go back 
                    if ( timerShowModalSet != undefined ) {
                            clearTimeout(timerShowModalSet)
                    }
                    level           -= 1
                    var parentCellId = '';
                    try {
                        graph.getModel().beginUpdate();
                        graph.removeCells(graph.getChildVertices(graph.getDefaultParent()));
                        var xml = previousGraphXml.pop();
                        var doc = mxUtils.parseXml(xml);
                        var codec = new mxCodec(doc);
                        var model = codec.decode(doc.documentElement);

                        // Find out the parent of the previous level to highlight it with some animation. 
                        // Example: if we go back to List_VPC level, we need to find the first VPC Cell and use its ParecenCellId
                        for (var key in model.cells)
                        {
                            var tmp = model.getCell(key);
                            if ( tmp.data != undefined && tmp.data.parentCellId != undefined ) {
                                if ( LEVELS[level] == LIST_VPC && tmp.data.awsResourceType == "vpc" ) {
                                    parentCellId = tmp.data.parentCellId;
                                    break;
                                } else
                                if ( LEVELS[level] == LIST_SUBNET && tmp.data.awsResourceType == "subnet" ) {
                                    parentCellId = tmp.data.parentCellId;
                                    break;
                                }
                            }
                        }
                        graph.getModel().mergeChildren(model.getRoot().getChildAt(0), graph.getDefaultParent());
                    } finally {
                        graph.getModel().endUpdate();
                    }
                    // Animation
                    var xs = [10,-10,-5,5,20,25];
                    var ys = [10,-10,-5,5,20,25];
                    var x  = xs[Math.floor(Math.random() * (5 - 0) + 0)];
                    var y  = ys[Math.floor(Math.random() * (5 - 0) + 0)];
                    var previousCoord;
                    setTimeout(function(){
                        previousCoord = moveMe(graph,parentCellId,x,y,true);
                    },70);
                    setTimeout(function(){
                        moveMe(graph,parentCellId,previousCoord.x,previousCoord.y, false);
                    },450);
                 }

                 function showModalInfo(currentState) {
                    // Check if the Mouse is over the same Object/Icon for a specific amount of time (set it on the calling of this throuhg setTimeOut, down below) 
                    // If the mouse has moved away after some time, the ModalInfo should not be shown, otherwise should be shown
                    var sameRegionYet = (globalCurrentState == currentState) && (globalCurrentState != undefined) && (globalCurrentState.cell != undefined);
                    if (!sameRegionYet)
                        return;

                    var divContent = document.getElementById(divContentId);
                    divContent.innerHTML  = currentState.cell.modalInfo.html;

                    if ( !showModalRightBtnFixed )
                         document.querySelector("#imgBtnClose").style.display = "none";

                    x = currentState.cell.getGeometry().x + (currentState.cell.getGeometry().width);
                    y = currentState.cell.getGeometry().y - 50;
                    var div = document.getElementById(divModalInfoId);
                    
                    var win = window,
                        doc = document,
                        docElem = doc.documentElement,
                        body    = doc.getElementsByTagName('body')[0],
                        width   = win.innerWidth || docElem.clientWidth || body.clientWidth,
                        height  = win.innerHeight|| docElem.clientHeight|| body.clientHeight;

                    div.style.display = "block";    
                        
                    // Adjust the Y position of the Div accordingly with the screen height    		 
                    divModalInfoHeight = div.offsetHeight;
                    if ( (y + divModalInfoHeight) > height ) {
                        y = y - ((y + divModalInfoHeight) - height) - (currentState.cell.getGeometry().height * 1.8); 
                    }
                    if ( y < 0 ) {
                        y = currentState.cell.getGeometry().y + 30;
                    }

                    // Adjust the X position of the Div accordingly with the screen width and the cell width
                    divModalInfoWidth = div.offsetWidth;
                    if ( (x + divModalInfoWidth) > width ) {
                        x = x - (divModalInfoWidth + currentState.cell.getGeometry().width + 5); 
                    }
                    
                    div.style.display = "block";
                    div.style.left    = x;
                    div.style.top     = y;
                 }

                 var delayTimeShowModalInfo = 500; // one second
                 var timerShowModalSet      = undefined;
                 var divModalInfoWidth      = 350;
                 var divModalInfoHeight     = 250;
                 var globalCurrentState     = null;
                 var showModalRightBtnFixed = false;
                 var divModalInfo           = document.createElement("div");   
                 var divModalInfoId         = "divModalInfo"
                 divModalInfo.style.display = "none";
                 divModalInfo.className     = "modalInfo";
                 //divModalInfo.innerHTML     = "Header";
                 divModalInfo.id            = divModalInfoId;
                 var divContentId           = "divContentId";
                 var divContent             = document.createElement("div");
                 divContent.innerHTML       = "Here, you can put any additional information";
                 divContent.id              = divContentId;
                 divModalInfo.appendChild(divContent);    

                 function main(container)
                 {
                     document.body.appendChild(divModalInfo);   

                     """ + self.htmlSnippets.jsBrowserNotSupported() + """
                     else
                     {
                         // Speedup the animation
                         mxText.prototype.enableBoundingBox = false
                         mxEvent.disableContextMenu(document.body);

                         // Creates the graph inside the given container
                         var graph = new mxGraph(container);
                         
                         // Disables all built-in interactions
                         graph.setEnabled(false);
                         graph.setPanning(true);
                         graph.panningHandler.useLeftButtonForPanning = true;
                         graph.setHtmlLabels(true);
         
                         // Handles clicks on cells
                         graph.addListener(mxEvent.CLICK, function(sender, evt)
                         {
                             var rightButton;
                             if ( evt.properties != undefined && evt.properties.event != undefined && evt.properties.event.button != undefined ) {
                                 rightButton = evt.properties.event.button == 2 ? true : false;
                             }
                             if ( rightButton != undefined && rightButton == false ) {
                                  var cell = evt.getProperty('cell');
                                  if (cell != null)
                                  {
                                      // If the Cell clicked is a VPC Resource and we are at the level of LIST_SUBNET, so...
                                      // we are clicking at the center of the graph, that is, the root parent (VPC) of all Subnets 
                                      // Then, we must go back to the previous Level: LIST_VPC(1) ---> AWS (0)
                                      if ( cell.data != undefined && cell.data.awsResourceType == "vpc" && LEVELS[level] == LIST_SUBNET ) {
                                          historyGraphGoBack(graph);
                                      } else
                                      if ( cell.data != undefined && cell.data.awsResourceType == "subnet" && LEVELS[level] == LIST_EC2 ) {
                                          historyGraphGoBack(graph);
                                      } else {
                                          // Before going foward, clean any ModalInfo TimeOut Set
                                          if ( timerShowModalSet != undefined ) {
                                              clearTimeout(timerShowModalSet)
                                          }
                                          // Save snapshot of the previous graph to go back when requested
                                          var enc          = new mxCodec();
                                          var node         = enc.encode(graph.getModel());
                                          var tmpXml       = mxUtils.getXml(node);
                                          if ( callAWSGraphNavigation(graph, cell) ) {
                                              // Only save when any call was implemented
                                              previousGraphXml.push(tmpXml);    
                                          }
                                      }
                                  }
                             } else {
                                 // Right Mouse Button Pressed
                                 showModalRightBtnFixed = true;
                                 showModalInfo(globalCurrentState);
                             }
                         });

                         // Handles mouseOver, mouseOut - ModalInfo
                         graph.addMouseListener({
                            currentState: null,
                            mouseMove: function(sender, evt) {
                                 if ( !showModalRightBtnFixed ) {
                                    if (this.currentState != null && (evt.getState() == this.currentState || evt.getState() == null) ) {
                                        globalCurrentState = this.currentState;
                                        if ( timerShowModalSet != undefined ) {
                                            clearTimeout(timerShowModalSet)
                                        }
                                        timerShowModalSet = setTimeout(showModalInfo,delayTimeShowModalInfo,this.currentState);
                                        if ( evt.getState() ) {
                                            evt.getState().setCursor('pointer');
                                        }     
                                    }

                                    var tmp = graph.view.getState(evt.getCell()); 
                                    
                                    if ( tmp != this.currentState ) {
                                        var div = document.getElementById(divModalInfoId);
                                        div.style.display = "none";
                                        if ( timerShowModalSet != undefined ) {
                                            clearTimeout(timerShowModalSet)
                                        }
                                        if ( tmp != null && tmp.cell.isVertex() ) {
                                            this.currentState  = tmp;
                                            globalCurrentState = tmp;
                                        }
                                    }
                                 }
                            },
                            mouseDown: function(sender, evt) { },
                            mouseUp:   function(sender, evt) { },
                            dragEnter: function(evt, state) { },
                            dragLeave: function(evt, state) { }
                         });

                         // Mouse Wheel Zooming
                         mxEvent.addMouseWheelListener((evt, up) => {
                            if (mxEvent.isConsumed(evt)) {
                                return;
                            }
                            let gridEnabled = graph.gridEnabled;
                            // disable snapping
                            graph.gridEnabled = false;
                            let p1 = graph.getPointForEvent(evt, false);
                            if (up) {
                                graph.zoomIn();
                            } else {
                                graph.zoomOut();
                            }
                            let p2 = graph.getPointForEvent(evt, false);
                            let deltaX = p2.x - p1.x;
                            let deltaY = p2.y - p1.y;
                            let view = graph.view;
                            view.setTranslate(view.translate.x + deltaX, view.translate.y + deltaY);
                            graph.gridEnabled = gridEnabled;
                            mxEvent.consume(evt);
                         }, container);
                          
                         // Changes the default vertex style in-place
                         var style = graph.getStylesheet().getDefaultVertexStyle();
                         // Rectangle (rounded)
                         style[mxConstants.STYLE_SHAPE] = mxConstants.SHAPE_RECTANGLE;
                         style[mxConstants.STYLE_ROUNDED] = true;
                         style[mxConstants.STYLE_PERIMETER] = mxPerimeter.RectanglePerimeter;
                         style[mxConstants.STYLE_GRADIENTCOLOR] = 'white';
                                         
                         // Gets the default parent for inserting new cells. This
                         // is normally the first child of the root (ie. layer 0).
                         var parent = graph.getDefaultParent();
         
                         var cx = graph.container.clientWidth / 2;
                         var cy = graph.container.clientHeight / 2;
                         
                         // Root of All
                         var awsImg = dataURIFor("aws-4.png").replace(";base64,",",");
                         var prefix = 'shape=image;image=' + awsImg + ";";

                         var cell       = graph.insertVertex(parent, 'AWS', 'AWS', cx - 20, cy - 15, 60, 40,prefix+'verticalLabelPosition=bottom;verticalAlign=top');
                         cell.data      = new CustomData('aws', 'root');
                         cell.modalInfo = new ModalInfoHTML("root",{"aws":"aws"});
                         
                         // Animates the changes in the graph model
                         graph.getModel().addListener(mxEvent.CHANGE, function(sender, evt)
                         {
                             var changes = evt.getProperty('edit').changes;
                             mxEffects.animateChanges(graph, changes);
                         });

                         """ +  self.htmlSnippets.jsMxToolBox("90","15") + """

                         var parameters = undefined;
                         postRequest(LIST_VPC,parameters,function(response) {
                             drillDown(graph, cell, response);
                         });
                     }
                 };

                 function closeModalInfo() {
                    var div = document.getElementById(divModalInfoId);
                    div.style.display      = "none";
                    showModalRightBtnFixed = false;
                 }

                 var LIST_VPC            = "listVpc";
                 var LIST_SUBNET         = "listSubnet";
                 var LIST_EC2            = "listEc2";
                 var LIST_SECURITY_GROUP = "listSecurityGroup";
                 var level               = 0;
                 var LEVELS              = ["ROOT", LIST_VPC, LIST_SUBNET, LIST_EC2];

                 function showLoading(show) {
                     if (show) {
                         document.querySelector("#loadingBackground").style.display = "";
                         document.querySelector("#loadingMessage").style.display = "";
                     } else {
                         document.querySelector("#loadingBackground").style.display = "none";
                         document.querySelector("#loadingMessage").style.display    = "none";
                     }
                 }

                 function callAWS(request, callback, event) {
                    showLoading(true);
                    postRequest(request.call,request.parameters,function(response, event) {
                        showLoading(false);
                        callback(response, event);
                    }, event);
                 }
                 function callListSecurityGroups(groupId, callback, event) {
                    var parameters  = [];
                    parameters.push({"name":"groupId","value":groupId});
                    var request = {
                        "call":LIST_SECURITY_GROUP,
                        "parameters": parameters
                    };
                    callAWS(request, callback, event);
                 }

                 function callAWSGraphNavigation(graph, cell) {
                     var callImplemented = false;
                     var awsResourceType = cell.data.awsResourceType;
                     if ( awsResourceType == "subnet" ) {
                         // Drill Down Subnets --> EC2s
                         showLoading(true);
                         callImplemented = true;
                         var parameters  = [];
                         parameters.push({"name":"vpcId","value":cell.data.awsResourceId},
                                         {"name":"subnetId","value":cell.data.awsResourceId});
                         postRequest(LIST_EC2,parameters,function(response) {
                             drillDown(graph, cell, response);
                             level = 3;
                         });
                     } else
                     if ( awsResourceType == "vpc" ) {
                         // Drill Down VPC --> Subnets
                         showLoading(true);
                         callImplemented = true;
                         var parameters  = [];
                         parameters.push({"name":"vpcId","value":cell.data.awsResourceId});
                         postRequest(LIST_SUBNET,parameters,function(response) {
                             drillDown(graph, cell, response);
                             level = 2;
                         });
                     } else
                     if ( awsResourceType == "aws" ) {
                         // Drill Down AWS --> VPCs
                         showLoading(true);
                         callImplemented = true;
                         var parameters  = undefined;
                         postRequest(LIST_VPC,parameters,function(response) {
                             drillDown(graph, cell, response);
                             level = 1;
                         });
                     }
                     return callImplemented;
                 }

                 function postRequest(action, parameters, callback, event) {
                     Http.open("POST", url, true);
                     Http.setRequestHeader('Content-type','application/json');
                     request = createRequest(action,parameters);
                     Http.onreadystatechange = (e) => {
                        if ( Http.readyState == 4 && Http.status == 200 ) {
                            callback(JSON.parse(Http.responseText), event)
                        } else {
                            //console.log("onreadystatechange = " + Http.status);
                        }
                     }
                     Http.onerror = (e) => {
                         showLoading(false);
                         alert("No connection available for " + url);
                     }
                     //Http.send(JSON.stringify(request));
                     Http.send(request);
                 }

                 function createRequest(action, params) {
                    var request        = {};
                    var parameters     = [];
                    request.action     = action;
                    request.parameters = parameters;

                    if ( params && params != undefined && params.length > 0 ) {
                        for (index = 0; index < params.length; ++index) {
                            request.parameters.push({
                                "name": params[index]["name"],
                                "value": params[index]["value"]
                            });
                        }
                    }
                    return JSON.stringify(request);
                }

                 // Loads the links for the given cell into the given graph
                 // by requesting the respective data in the server-side
                 // (implemented for this demo using the server-function)
                 function drillDown(graph, cell, json)
                 {
                     if (graph.getModel().isVertex(cell))
                     {
                         centerCell = cell;
                         var cx = graph.container.clientWidth / 2;
                         var cy = graph.container.clientHeight / 2;
                         
                         // Gets the default parent for inserting new cells. This
                         // is normally the first child of the root (ie. layer 0).
                         var parent = graph.getDefaultParent();
         
                         // Adds cells to the model in a single step
                         graph.getModel().beginUpdate();
                         try
                         {
                             var xml = jsonToGraphModel(cell.id, json);
                             var doc = mxUtils.parseXml(xml);
                             var dec = new mxCodec(doc);
                             var model = dec.decode(doc.documentElement);
         
                             // Removes all cells which are not in the response
                             for (var key in graph.getModel().cells)
                             {
                                 var tmp = graph.getModel().getCell(key);
                                 
                                 if (tmp != cell &&
                                     graph.getModel().isVertex(tmp))
                                 {
                                     graph.removeCells([tmp]);
                                 }
                             }
         
                             // Merges the response model with the client model
                             graph.getModel().mergeChildren(model.getRoot().getChildAt(0), parent);
         
                             // Moves the given cell to the center
                             var geo = graph.getModel().getGeometry(cell);
         
                             if (geo != null)
                             {
                                 geo = geo.clone();
                                 geo.x = cx - geo.width / 2;
                                 geo.y = cy - geo.height / 2;
         
                                 graph.getModel().setGeometry(cell, geo);
                             }
         
                             // Creates a list of the new vertices, if there is more
                             // than the center vertex which might have existed
                             // previously, then this needs to be changed to analyze
                             // the target model before calling mergeChildren above
                             var vertices = [];
                             
                             for (var key in graph.getModel().cells)
                             {
                                 var tmp = graph.getModel().getCell(key);
                                 
                                 if (tmp != cell && model.isVertex(tmp))
                                 {
                                     vertices.push(tmp);
         
                                     // Changes the initial location "in-place"
                                     // to get a nice animation effect from the
                                     // center to the radius of the circle
                                     var geo = model.getGeometry(tmp);
         
                                     if (geo != null)
                                     {
                                         geo.x = cx - geo.width / 2;
                                         geo.y = cy - geo.height / 2;
                                     }
                                 }
                             }
                             
                             // Arranges the response in a circle
                             var cellCount = vertices.length;
                             var phi = 2 * Math.PI / cellCount;
                             var r = Math.min(graph.container.clientWidth / 2.3,
                                              graph.container.clientHeight / 2.3);
                             
                             for (var i = 0; i < cellCount; i++)
                             {
                                 var geo = graph.getModel().getGeometry(vertices[i]);
                                 
                                 if (geo != null)
                                 {
                                     geo = geo.clone();
                                     geo.x += r * Math.sin(i * phi);
                                     geo.y += r * Math.cos(i * phi);
         
                                     graph.getModel().setGeometry(vertices[i], geo);
                                 }
                             }
                         }
                         finally
                         {
                             // Updates the display
                             graph.getModel().endUpdate();
                         }
                     }
                     showLoading(false);
                 };

                 
                 var colors      = ["rgb(0,0,255)","rgb(183,151,43)","rgb(110,105,89)","rgb(91,154,176)","rgb(255,0,9)","rgb(204,204,0)","rgb(128,128,128)","rgb(246,0,0)","rgb(246,178,103)","rgb(8,180,151)"];
                 var labelColors = {};
                 function colorForLabel(label) {
                    var color; 
                    if (labelColors[label] && labelColors[label] != undefined && labelColors[label] != "") {
                        color = labelColors[label];
                    }
                    if ( !color || color == undefined || color == "" ) {
                        var rnd = Math.floor(Math.random() * colors.length-1);
                        if ( rnd < 0 ) {
                            rnd = 3;
                        }
                        color = colors[rnd];
                        labelColors[label] = color;
                    }
                    return color;
                 }
                 function componentToHex(c) {
                    var hex = c.toString(16);
                    return hex.length == 1 ? "0" + hex : hex;
                 }
                 function rgbToHex(r, g, b) {
                    return "#" + componentToHex(r) + componentToHex(g) + componentToHex(b);
                 }
                 function rgbToHexStr(str) {
                    var cleanRGB = (str.replace("rgb(","").replace(")","")).split(","); 
                    return rgbToHex(parseInt(cleanRGB[0]), parseInt(cleanRGB[1]), parseInt(cleanRGB[2]));
                 }

                 function jsonToGraphModel(cellId, json)
                 {
                     // Creates a local graph with no display
                     var graph = new mxGraph();
                     
                     // Gets the default parent for inserting new cells. This
                     // is normally the first child of the root (ie. layer 0).
                     var parent = graph.getDefaultParent();
         
                     // Adds cells to the model in a single step
                     graph.getModel().beginUpdate();
                     try
                     {
                        if ( json != undefined ) {
                            action = json["aws"]["action"]

                            width  = 130;
                            height = 50;
                            
                            if (action == LIST_VPC) {
                                var v0  = graph.insertVertex(parent, cellId, 'AWS-Dummy', 0, 0, width, height);
                                v0.data    = new CustomData('root', undefined, json, '');
                                var prefix = 'shape=image;image=' + dataURIFor(AWS_VPC_IMG) + ";";
                                for( var key in json["aws"]["vpc"] ) {
                                    var vpc     = json["aws"]["vpc"][key];
                                    var id      = vpc.vpcId;
                                    var label   = "<b>" + vpc.vpcId + "</b><br>" + vpc.cidrBlock; 
                                    var v       = graph.insertVertex(parent, id, label, 0, 0, width, height, prefix+'verticalLabelPosition=bottom;verticalAlign=top;whiteSpace=wrap;');
                                    v.data      = new CustomData('vpc', vpc.vpcId, json, cellId);
                                    v.modalInfo = new ModalInfoHTML('vpc',vpc);
                                    var e       = graph.insertEdge(parent, null, '', v0, v);
                                }
                            } else
                            if (action == LIST_SUBNET) {
                                 var v0      = graph.insertVertex(parent, cellId, 'Vpc', 0, 0, width, height);
                                 v0.data     = new CustomData('vpc', undefined, json);
                                 var prefix1 = 'shape=image;image=' + dataURIFor(AWS_SUBNET_PRIVATE_IMG) + ";";
                                 var prefix2 = 'shape=image;image=' + dataURIFor(AWS_SUBNET_PUBLIC_IMG) + ";";
                                 var prefix  = prefix1;
                                 for( var key in json["aws"]["vpc"]["subnets"] ) {
                                    var subnet = json["aws"]["vpc"]["subnets"][key];
                                    var id           = subnet.subnetId;
                                    var defaultForAz = subnet.defaultForAz ? "<br>DEFAULT" : "";
                                    if (subnet.scope == "public") {
                                        prefix = prefix2
                                    } else {
                                        prefix = prefix1
                                    }
                                    var colorAz = colorForLabel(subnet.availabilitZone);
                                    var label   = "<b>" + subnet.subnetId + "</b><br>" + subnet.cidrBlock + "<br><span style='color:" + colorAz + ";'>" + subnet.availabilitZone + "</span>" + defaultForAz;
                                    var v       = graph.insertVertex(parent, id, label, 0, 0, width, height, prefix+'verticalLabelPosition=bottom;verticalAlign=top;whiteSpace=wrap;');
                                    v.data      = new CustomData('subnet', subnet.subnetId, json, cellId);
                                    v.modalInfo = new ModalInfoHTML('subnet',subnet);
                                    var e       = graph.insertEdge(parent, null,subnet.availabilitZone, v0, v,'strokeColor=' + rgbToHexStr(colorAz) + ';');
                                }
                            } else
                            if (action == LIST_EC2) {
                                 var v0     = graph.insertVertex(parent, cellId, 'Subnet', 0, 0, width, height);
                                 v0.data    = new CustomData('ec2', undefined, json);
                                 for( var key in json["aws"]["ec2"]["instances"] ) {
                                    var ec2instance = json["aws"]["ec2"]["instances"][key];

                                    var prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_IMG) + ";";
                                    if ( ec2instance.instanceType.startsWith("t2") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_T2_IMG) + ";";
                                    } else 
                                    if ( ec2instance.instanceType.startsWith("t3a") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_T3A_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("t3") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_T3_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("a1") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_A1_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("c4") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_C4_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("c5") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_C5_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("m4") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_M4_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("m5a") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_M5A_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("m5") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_M5_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("p2") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_P2_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("p3") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_P3_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("r4") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_R4_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("r5a") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_R5A_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("r5") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_R5_IMG) + ";";
                                    }
                                    
                                    var id           = ec2instance.instanceId;
                                    var colorInsType = colorForLabel(ec2instance.instanceType);
                                    var colorState   = colorForLabel(ec2instance.state.name.toUpperCase());
                                    var label        = "<b>" + ec2instance.instanceId + "</b><br>" + ec2instance.privateIpAddress + "<br><span style='color:" + colorInsType + ";'>" + ec2instance.instanceType + "</span>";
                                    var v            = graph.insertVertex(parent, id, label, 0, 0, width, height, prefix+'verticalLabelPosition=bottom;verticalAlign=top;whiteSpace=wrap;');
                                    v.data           = new CustomData('ec2', ec2instance.instanceId, json, v0.cellId);
                                    v.modalInfo      = new ModalInfoHTML('ec2',ec2instance);
                                    var e            = graph.insertEdge(parent,null,ec2instance.state.name.toUpperCase(), v0, v, 'strokeColor=' + rgbToHexStr(colorState) + ';');
                                }
                            }
                        } else {
                            console.log("json is not defined");
                        }
                     }
                     finally
                     {
                         // Updates the display
                         graph.getModel().endUpdate();
                     }
         
                     var enc = new mxCodec();
                     var node = enc.encode(graph.getModel());
                     
                     return mxUtils.getXml(node);
                 };

                 function CustomData(awsResourceType, awsResourceId, sourceData, parentCellId) {
                    this.awsResourceType = awsResourceType;
                    this.awsResourceId   = awsResourceId;
                    this.sourceData      = sourceData;
                    this.parentCellId    = parentCellId;
                 }

                 function ModalInfoHTML(resourceId, data) {
                    if ( resourceId == 'root' ) {
                       this.html = buildHtmlAwsModalInfo(data); 
                    } else
                    if ( resourceId == 'vpc' ) {
                       this.html = buildHtmlVpcModalInfo(data);
                    } else
                    if ( resourceId == 'subnet' ) {
                       this.html = buildHtmlSubnetModalInfo(data);   
                    } else
                    if ( resourceId == 'ec2' ) {
                       this.html = buildHtmlEc2ModalInfo(data);
                    } else {
                       this.html = "<h5>" + JSON.stringify(data) + "</h5>"; 
                    }
                 }


                 var tagBehavior = 'onmouseover="javascript:this.style.backgroundColor=' + "'" + 'rgb(0,175,175)' + "'" + ';this.style.cursor=' + "'" + 'pointer' + "'" + ';"onmouseout="javascript:this.style.backgroundColor=' + "'" + 'rgb(0,220,220)' + "';" + '"';
                 function buildHtmlVpcModalInfo(data) {
                     var html =
                         '<div id="imgBtnClose" style="position:absolute;top:7%;left:93%"><img  onMouseOver="this.style.cursor=' + "'" + 'pointer' + "'" + '" src=""" + self.images.close +""" onClick="javascript:closeModalInfo();"></div>'
                       + '<table class="tableInfo" width="100%" height="100%" border="0" cellpadding="0" cellspacing="0">'
                       + ' <tr><td align="center" height="35px" colspan="2" class="cellHeader">VPC</td></tr>'  
                       + ' <tr><td class="cellInfo">'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.vpcId + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     Vpc Id.......:&nbsp;<font style="color:blue;">' + data.vpcId + '</font><br>'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.cidrBlock + "'"+ ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     CIDR Block...:&nbsp;<font style="color:blue;">' + data.cidrBlock + '</font>'
                       + ' <tr>'
                       + ' <tr><td colspan="2" class="cellTags">'
                       + '     Tags:<br><br>'
                       var tagsPerLine = 0;
                       for(var index = 0; index < data.tags.length; index++) {
                           var tag = data.tags[index];
                           html = html + '<div class="tag"' 
                                       + 'onclick="javascript:copyToClipboard(' + "'" + tag.key + "=" + tag.value + "'" + ');" '
                                       + tagBehavior + '>' + tag.key + '=' + tag.value + '</div>';
                           tagsPerLine += 1 
                           if (tagsPerLine >= 2) {
                               tagsPerLine = 0;
                               html = html + '<br><br>'
                           }
                       }
                       + ' </td>'
                       + '</table>'
                    return html;  
                 }

                 function buildHtmlSubnetModalInfo(data) {
                     var labelScope = data.scope == "public" ? "PUBLIC" : "PRIVATE";
                     var html =  
                         '<div id="imgBtnClose" style="position:absolute;top:5%;left:93%"><img  onMouseOver="this.style.cursor=' + "'" + 'pointer' + "'" + '" src=""" + self.images.close +""" onClick="javascript:closeModalInfo();"></div>'
                       + '<table class="tableInfo" width="100%" height="100%" border="0" cellpadding="0" cellspacing="0">'
                       + ' <tr><td align="center" height="35px" colspan="2" class="cellHeader">' + labelScope + ' SUBNET</td></tr>'  
                       + ' <tr><td class="cellInfo">'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.subnetId + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     Subnet Id.......:&nbsp;<font style="color:blue;">' + data.subnetId + '</font><br>'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.availabilitZone + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     AZ..............:&nbsp;<font style="color:blue;">' + data.availabilitZone + ' (' + data.availabilitZoneId  + ')' + '</font><br>'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.cidrBlock + "'"+ ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     CIDR Block......:&nbsp;<font style="color:blue;">' + data.cidrBlock + '</font><br>'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.availableIpAddressCount + "'"+ ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     Available IpV4..:&nbsp;<font style="color:blue;">' + data.availableIpAddressCount + '</font><br>'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.vpcId + "'"+ ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     VPC Id..........:&nbsp;<font style="color:blue;">' + data.vpcId + '</font><br>'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.defaultForAz + "'"+ ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     Default for AZ..:&nbsp;<font style="color:blue;">' + data.defaultForAz + '</font><br>'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.subnetArn + "'"+ ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     Subnet ARN......:&nbsp;<font style="color:blue;font-size:9px;">' + data.subnetArn + '</font>'
                       + ' <tr>'
                       + ' <tr><td colspan="2" class="cellTags">'
                       + '     Tags:<br><br>'
                       var tagsPerLine = 0;
                       for(var index = 0; index < data.tags.length; index++) {
                           var tag = data.tags[index];
                           html = html + '<div class="tag"' 
                                       + 'onclick="javascript:copyToClipboard(' + "'" + tag.key + "=" + tag.value + "'" + ');" '
                                       + tagBehavior + '>' + tag.key + '=' + tag.value + '</div>';
                           tagsPerLine += 1 
                           if (tagsPerLine >= 2) {
                               tagsPerLine = 0;
                               html = html + '<br><br>'
                           }
                       }
                       + ' </td>'
                       + '</table>'
                    return html;  
                 }

                 function thousandSeparator(x) {
                    return x.toString().replace(/\B(?=(\d{3})+(?!\d))/g, ".");
                 }

                 function callBackShowHtmlSecurityGroup(response, event) {
                     //TODO: Build the HTML Popup for Security Group
                     console.log(event);
                     console.log(response);
                 }

                 function buildHtmlEc2ModalInfo(data) {
                     var colorState;
                     if ( data.state.name.toUpperCase() == "RUNNING" ) {
                         colorState = "green";
                     } else
                     if ( data.state.name.toUpperCase() == "STOPPED" ) {
                         colorState = "red";
                     } else {
                         colorState = "black";
                     }
                     var dtToday      = new Date();
                     var dt           = data.launchTime.split("-")[1] + "-" + data.launchTime.split("-")[0] + "-" + data.launchTime.split("-")[2];
                     var dtLaunched   = new Date(dt);
                     var timeLaucnhed = milliseconds = Math.abs(dtLaunched - dtToday);
                     var hours        = "&nbsp;(<font color=gray;>" + thousandSeparator(parseInt((milliseconds / 36e5))) + " hours)</font>";
                     var publicIpAddressEntry = "";
                     if ( data.publicIpAddress && data.publicIpAddress != "" ){
                           publicIpAddressEntry = '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.publicIpAddress + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                                                + '     Public IP...........:&nbsp;<font style="color:blue;">' + data.publicIpAddress + '</font><br>';
                     }
                     var blockDevices = "";
                     if ( data.blockDeviceMappings ) {
                          var totalDevices = data.blockDeviceMappings.length
                          for( var i = 0; i < data.blockDeviceMappings.length; i++ ) {
                              var blockDevice = data.blockDeviceMappings[0];
                              var labelTotal  = '<font style="color:gray;">[' + (i+1) + ' of ' + totalDevices + ']</font>';
                              blockDevices =  blockDevices
                              + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + blockDevice.ebs.volumeId + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                              + '     EBS...............' + labelTotal + '..:&nbsp;<font style="color:blue;">' + blockDevice.deviceName + ' (' + blockDevice.ebs.volumeId +')&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</font><br>'
                              + '&nbsp;&nbsp;&nbsp;&nbsp;|Status....................:&nbsp;<font style="color:blue;">' + blockDevice.ebs.status + '</font><br>'
                              + '&nbsp;&nbsp;&nbsp;&nbsp;|Attach Time...............:&nbsp;<font style="color:blue;">' + blockDevice.ebs.attachTime + '</font><br>'
                              + '&nbsp;&nbsp;&nbsp;&nbsp;|Delete on Terminatation...:&nbsp;<font style="color:blue;">' + blockDevice.ebs.deleteOnTermination + '</font><br>'  
                          }
                     }
                     var securityGroups = "";
                     if ( data.securityGroups ) {
                          var totalSecurityGroups = data.securityGroups.length
                          for( var i = 0; i < data.securityGroups.length; i++ ) {
                              var securityGroup = data.securityGroups[0];
                              var labelTotal  = '<font style="color:gray;">[' + (i+1) + ' of ' + totalSecurityGroups + ']</font>';
                              securityGroups =  securityGroups
                              + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + securityGroup.groupId + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                              + '     Security Group....' + labelTotal + '..:&nbsp;<font style="color:darkorange;text-decoration-line:underline;">'
                              + '<span onMouseOver="this.style.cursor=' + "'" + 'pointer' + "'" + '" onClick="javascript:callListSecurityGroups(' + "'" + securityGroup.groupId + "'" + ',callBackShowHtmlSecurityGroup, event);">' + securityGroup.groupId + '</span></font><br>'
                              + '&nbsp;&nbsp;&nbsp;&nbsp;|Group Name................:&nbsp;<font style="color:blue;">' + securityGroup.groupName + '</font><br>'
                          }
                     }
                     var html =  
                         '<div id="imgBtnClose" style="position:absolute;top:4%;left:93%"><img  onMouseOver="this.style.cursor=' + "'" + 'pointer' + "'" + '" src=""" + self.images.close +""" onClick="javascript:closeModalInfo();"></div>'
                       + '<table class="tableInfo" width="100%" height="100%" border="0" cellpadding="0" cellspacing="0">'
                       + ' <tr><td align="center" height="35px" colspan="2" class="cellHeader">EC2 ' + data.instanceType.toUpperCase() + '</td></tr>'  
                       + ' <tr><td class="cellInfo">'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.instanceId + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     Instance Id.................:&nbsp;<font style="color:blue;">' + data.instanceId + '</font><br>'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.privateIpAddress + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     Private IP..................:&nbsp;<font style="color:blue;">' + data.privateIpAddress + '</font><br>'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.launchTime + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     Launch Time.................:&nbsp;<font style="color:blue;">' + data.launchTime + '</font>' + hours + '<br>'
                       + publicIpAddressEntry
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.state.name.toUpperCase() + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     State.......................:&nbsp;<font style="color:' + colorState +';">' + data.state.name.toUpperCase() + '</font><br>'

                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.cpu.coreCount + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     CPU Core Count..............:&nbsp;<font style="color:blue;">' + data.cpu.coreCount + '</font><br>'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.cpu.coreCount + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     CPU Threads Per Core........:&nbsp;<font style="color:blue;">' + data.cpu.threadsPerCore + '</font><br>'
                       + securityGroups
                       + blockDevices
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.subnetId + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     Subnet Id...................:&nbsp;<font style="color:blue;">' + data.subnetId + '</font><br>'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.vpcId + "'"+ ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     VPC Id......................:&nbsp;<font style="color:blue;">' + data.vpcId + '</font><br>'
                       + ' <tr>'
                       + ' <tr><td colspan="2" class="cellTags">'
                       + '     Tags:<br><br>'
                       var tagsPerLine = 0;
                       for(var index = 0; index < data.tags.length; index++) {
                           var tag = data.tags[index];
                           html = html + '<div class="tag"' 
                                       + 'onclick="javascript:copyToClipboard(' + "'" + tag.key + "=" + tag.value + "'" + ');" '
                                       + tagBehavior + '>' + tag.key + '=' + tag.value + '</div>';
                           tagsPerLine += 1 
                           if (tagsPerLine >= 2) {
                               tagsPerLine = 0;
                               html = html + '<br><br>'
                           }
                       }
                       + ' </td>'
                       + '</table>'
                    return html;  
                 }

                 function buildHtmlAwsModalInfo(data) {
                     var html =  
                         '<table class="tableInfo" width="100%" height="100%" border="0" cellpadding="0" cellspacing="0">'
                       + ' <tr><td align="center" height="35px" colspan="2" class="cellHeader">AWS</td></tr>'  
                       + ' <tr><td class="cellInfo">'
                       + '    <font style="color:blue;">&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;AWS&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</font><br>'
                       + ' <tr>'
                       + '</table>'
                    return html;  
                 }

                 function copyToClipboard(str) {
                    const el = document.createElement('textarea');
                    el.value = str;
                    document.body.appendChild(el);
                    el.select();
                    document.execCommand('copy');
                    document.body.removeChild(el);
                 }

                 """ + self.htmlSnippets.jsFormatXml() + """
                 function pad2(n) { return n < 10 ? '0' + n : n }

                 function timestamp() {
                    var date = new Date();
                    return  date.getFullYear().toString() + "/" + pad2(date.getMonth() + 1) + "/" +  pad2( date.getDate()) 
                    +  " " + pad2( date.getHours() ) + ":" + pad2( date.getMinutes() ) + ":" + pad2( date.getSeconds() ) + ":" + pad2( date.getMilliseconds() ) + " ";
                 }
             </script>
          </head>   
          """ + self.htmlSnippets.htmlCommonBody() + """
        </html>
        """
        self.openHTML(html)
        self.startHttpServer()
        print(Emoticons.ok() + " \033[93mNavigator exited!\033[0m")    
